import types

from Acquisition import aq_base
from AccessControl import ModuleSecurityInfo
from Products.PythonScripts.Utility import allow_class

from Products.Silva import adapters as silva_adapters
from Products.Silva import SilvaObject, Ghost, Folder, GhostFolder, Link, Asset

from utils import xml_reference_helper, default_check_map
import utils

from urlparse import urlparse

class ReferenceInfo:
    """simple record to contain information about one reference
    this class is horribly inflexible, but it works for now ...
    """

    # codes for the reference type

    # unknown / unclassified link
    REF_UNKNOWN = 0 
    # an internal link to something
    REF_INTERN = 1
    # an internal reference to something
    REF_MAILTO = 2
    # an external link
    REF_EXTERN = 3

    REF = {
        REF_UNKNOWN: 'Unknown',
        REF_INTERN: 'Intern',
        REF_MAILTO: 'Mailto:',
        REF_EXTERN: 'Extern'
        }


    # codes for restrictions
    RESTRICT_NONE=0
    RESTRICT_SILVA_OBJECT = 1
    RESTRICT_IMAGE = 2
    RESTRICT_FILE = 3
    RESTRICT_CONTENT = 4
    RESTRICT_FOLDER = 5
    # hmmm, more ?
    RESTRICT_CODE = 6
    RESTRICT_EXTSOURCE = 7

    RESTRICT = {
        RESTRICT_NONE: 'None',
        RESTRICT_SILVA_OBJECT: 'Silva Object',
        RESTRICT_IMAGE: 'Silva Image',
        RESTRICT_FILE: 'Silva File',
        RESTRICT_CONTENT: 'Silva Content',
        RESTRICT_FOLDER: 'Silva Folder',
        RESTRICT_CODE: 'Code Source',
        RESTRICT_EXTSOURCE: 'Silva External Source',
        }


    # information for path completion
    # (internal references only)
    
    # to be determined; checker will use default
    NO_PATH = 0
    # resolve path from zope root
    PATH_ZOPE_ROOT = 1
    # resolve path from silva root
    PATH_SILVA_ROOT = 2
    # resolve path from outmost container of containing object
    PATH_CONTAINER = 3
    # resolve path from containing object
    PATH_OBJECT = 4
    # resolve from the container _this_ object is located in
    # XXX this is ill defined yet. it has been invented
    # with the reference checker to solve problems with
    # virtual host settings; needs review
    PATH_THIS_CONTAINER = 5

    PATH = {
        NO_PATH: 'Undetermined',
        PATH_ZOPE_ROOT: 'Zope Root',
        PATH_SILVA_ROOT: 'Silva Root',
        PATH_CONTAINER: 'Current Container',
        PATH_OBJECT: 'Current Object',
        PATH_THIS_CONTAINER: 'This Container',
        }

    # status codes of the reference

    # unclassified
    STATUS_UNKNOWN = -1
    # ok
    STATUS_OK = 0
    # reference missing
    STATUS_MISSING = 1
    # reference exists but has no public version
    STATUS_UNPUBLISHED = 2
    # anchor in reference missing
    STATUS_MISSING_ANCHOR = 12
    # intern reference points to wrong type
    STATUS_WRONG_TYPE = 3
    # weird acquisition: the reference points to an object
    # in a non-straigth way
    STATUS_WEIRD_ACQUISITION = 4
    # external link has timeout
    STATUS_TIMEOUT = 6
    # external links which result in a error code
    STATUS_BROKEN = 7
    # external links which result in a non-200 code (e.g. redirects)
    STATUS_WARNING = 8

    # unchecked by config
    STATUS_UNCHECKED = 10

    STATUS = {
        STATUS_UNKNOWN: 'Unknown',
        STATUS_OK: 'OK',
        STATUS_MISSING: 'Missing',
        STATUS_WRONG_TYPE: 'Wrong type',
        STATUS_MISSING_ANCHOR: 'Missing Anchor',
        STATUS_WEIRD_ACQUISITION: 'Weird',
        STATUS_TIMEOUT: 'Timeout',
        STATUS_UNCHECKED: 'Unchecked',
        STATUS_UNPUBLISHED: 'Unpublished',
        STATUS_BROKEN: 'Error', # FIXME: wrong code        
        STATUS_WARNING: 'Warning', # FIXME: better message?
        }

    # hm :(
    __allow_access_to_unprotected_subobjects__=1

    def __init__(self, obj_path, target_path, ref_type=REF_UNKNOWN, status=STATUS_UNKNOWN,
                 restriction=RESTRICT_NONE, path_info=NO_PATH):
        if type(obj_path) != type( () ):
            obj_path = obj_path.getPhysicalPath()
        self.object = obj_path
        self.target = target_path # usually a string, but sometimes a tuple of ids (_content_path of Ghosts)
        self.ref_type = ref_type
        self.restriction = restriction
        self.status = status
        self.path_info = path_info
        self.message = ''

    def get_target(self, withAnchor=1):
        """Return target as string. Remove anchor if desired."""
        target = self.target
        if type(target) == types.TupleType or type(target) == types.ListType:
            target =  '/'.join(self.target)
        if type(target) == types.UnicodeType:
            # XXX what if we dont have ascii here?
            # should we flag warning? if yes, how?
            target = target.encode('ascii', 'replace')
        if not withAnchor:
            #try:
            #    # remove anchor from target
            #    res = target.split('#')
            #    target = res[0]
            #except: pass

            # this actually scraps params and query, too
            target = urlparse(target)[2]
        return target
        

    def set_message(self, msg):
        self.message=msg

    def set_status(self, status):
        self.status=status

    def refinfo(self):
        "onestring refinfo for debugging"
        return "type=%s restrict=%s path=%s status=%s" %\
               (self.REF.get(self.ref_type),
                self.RESTRICT.get(self.restriction),
                self.PATH.get(self.path_info),
                self.STATUS.get(self.status)
                )
    
    def get_problem_description(self):
        """used by the user interface for getting the problem description"""
        return self.STATUS.get(self.status,'[undefined]')


    def __str__(self):
        return "<RefInfo: %s object=%s target=%s>" % \
               (self.refinfo(),
                self.object,
                self.get_target() or '[[empty]]')
    
    __repr__ = __str__

    # define equality for the tests
    def __eq__(self, other):
        return self.object == other.object and \
               self.get_target() == other.get_target() and \
               self.ref_type == other.ref_type and \
               self.restriction == other.restriction and \
               self.status == other.status and \
               self.path_info == other.path_info

    def __ne__(self, other):
        return not self.__eq__(other)

    # no __hash__ : should not be used as dictionary key either ...

# ARGL: avoid recursive references:
utils.ReferenceInfo = ReferenceInfo

class BaseRetriever(silva_adapters.adapter.Adapter):
    " base class, also used as dummy for unknown cases"
    def __init__(self, object):
        self.object = object

    def get_references(self):
        return ()
        # XXX use this if debugging for missing adapters
        # (note that this makes the unit tests fail)
        return [ ReferenceInfo(self.object,
                               '/foo' + '  ' + self.object.meta_type,
                               ref_type=ReferenceInfo.REF_INTERN,
                               path_info=ReferenceInfo.PATH_SILVA_ROOT) ]


class NoneRetriever(BaseRetriever):

    def get_references(self):
        return ()    
    
class GhostRetriever(BaseRetriever):

    def get_references(self):
        return [ ReferenceInfo(self.object, self.object._content_path,
                               ref_type=ReferenceInfo.REF_INTERN,
                               path_info=ReferenceInfo.PATH_ZOPE_ROOT,
                               restriction=ReferenceInfo.RESTRICT_CONTENT),
                 ]

class GhostFolderRetriever(BaseRetriever):

    def get_references(self):
        return [ ReferenceInfo(self.object, self.object._content_path,
                               ref_type=ReferenceInfo.REF_INTERN,
                               path_info=ReferenceInfo.PATH_ZOPE_ROOT,
                               restriction=ReferenceInfo.RESTRICT_FOLDER),
                 ]

class LinkRetriever(BaseRetriever):

    def get_references(self):
        # XXX relative Link's, hmmm
        if self.object._link_type == 'relative':
            ref =  ReferenceInfo(self.object, self.object.get_url(),
                                   ref_type=ReferenceInfo.REF_INTERN,
                                   path_info=ReferenceInfo.PATH_CONTAINER,
                                   restriction=ReferenceInfo.RESTRICT_NONE),
        else:
            ref = ReferenceInfo(self.object,  self.object.get_url())
        return [ ref ]



class SQLDataSourceRetriever(BaseRetriever):
    
  def get_references(self):
    # check reference by ourselves, as the reference-checker
    # is not flexible enough :(
    status = ReferenceInfo.STATUS_OK
    connection_id = str(self.object.connection_id())
    try:
        o = getattr(self.object, connection_id)
        o = aq_base(o)
        # check if this is really a sql connection ... hmmm
        # cut & paste from the ZSQLMethods product :-/
        if hasattr(o,'_isAnSQLConnection') and o._isAnSQLConnection \
           and hasattr(o,'id'):
            # ok, if seems to be a connection
            pass
        else:
            status = ReferenceInfo.STATUS_WRONG_TYPE
    except (AttributeError, KeyError):
        status = ReferenceInfo.STATUS_MISSING

    return [ ReferenceInfo(self.object, connection_id,
                               ref_type=ReferenceInfo.REF_INTERN,
                               path_info=ReferenceInfo.PATH_OBJECT,
                               status = status) ]


class DocumentRetriever(BaseRetriever):
    def get_references(self):
        refs = []
        xml_reference_helper(refs, self.object.content.documentElement, \
                             default_check_map)
        return refs

# a simple map meta_type -> harvester class for now ...
retriever_registry = {}

retriever_registry[SilvaObject.SilvaObject] = BaseRetriever
retriever_registry[Folder.Folder] = NoneRetriever
retriever_registry[Asset.Asset] = NoneRetriever

retriever_registry[Link.LinkVersion] = LinkRetriever

retriever_registry[Ghost.GhostVersion] = GhostRetriever
retriever_registry[GhostFolder.GhostFolder] = GhostFolderRetriever


try:
    from Products.SilvaDocument.Document import DocumentVersion
    retriever_registry[DocumentVersion] = DocumentRetriever
except ImportError:
    try:
        from Products.Silva.Document import DocumentVersion
        retriever_registry[DocumentVersion] = DocumentRetriever
    except ImportError:
        pass

try:
    from Products.Silva.DemoObject import DemoObjectVersion
    retriever_registry[DemoObjectVersion] = DocumentRetriever
except ImportError:
    pass

try:
    from Products.SilvaDemoExtension.DemoExtensionObject import DemoExtensionObjectVersion
    retriever_registry[DemoExtensionObjectVersion] = DocumentRetriever
except ImportError:
    pass

try:
    from Products.SilvaExternalSources.SQLSource import SQLSource
    retriever_registry[SQLSource] = SQLDataSourceRetriever
except ImportError:
    pass

try:
    from Products.Silva.SQLDataSource import SQLDataSource
    retriever_registry[SQLDataSource] = SQLDataSourceRetriever
except ImportError:
    pass

    
def _getRetriever_forKlass(klass):
    # XXX argl: we mimic the python class attribute lookup rules here :-/
    # this may be braindead slow
    # actually it could already get unit tests by its own
    harv = retriever_registry.get(klass)
    if harv is None:
        for parent in klass.__bases__:
            h2 = _getRetriever_forKlass(parent)
            if h2 is not None:
                return h2
    return harv

def getReferenceRetriever(context):
    klass = _getRetriever_forKlass(aq_base(context).__class__)
    if klass is None:
        klass = BaseRetriever


    if not issubclass(klass, BaseRetriever):
        raise TypeError,'%s registered for %s is not a reference harvester'%\
              (klass, context.__class__)
    return klass(context)


allow_class(ReferenceInfo)

module_security = ModuleSecurityInfo('Products.SilvaReferenceChecker.adapters.reference_info')

# FIXME should read "declareProcteded(...)"
module_security.declarePublic('getReferenceRetriever')
