# python
from types import TupleType, ListType

# Zope
from Acquisition import aq_base
import zLOG

# Silva
from Products.Silva import SilvaObject, Ghost, Folder, Asset
from Products.Silva import SilvaPermissions

# Silva Reference Checker
from ReferenceChecker import ReferenceInfo


def get_version_for_state(object, state):
    # return the "right" version of the object for a given state
    # may return none, if the version does not exist
    # assumes state = "public" by default

    if state == 'editable':
        return object.get_editable()
    if state == 'previewable':
        return object.get_previewable()
    return object.get_viewable()


# a dummy method returning a test reference
# use this as monkey or SilvaObject to check if Your extensions
# have implemented the "get_reference" method;
# if instances of some content types show up with this
# reference in the reference check, they didn't
def get_references_dummy(self,version=None):
    return [ ReferenceInfo(self.getPhysicalPath(), '/foo' + '  ' + self.meta_type,
                           ref_type=ReferenceInfo.REF_INTERN,
                           path_info=ReferenceInfo.PATH_SILVA_ROOT) ]


def get_references_nothing(self, version=None):
    return ()

#SilvaObject.SilvaObject.get_references = get_references_dummy
SilvaObject.SilvaObject.get_references = get_references_nothing
#SilvaObject.SilvaObject.security.declareProtected(SilvaPermissions.ChangeSilvaViewRegistry,
#                                      'get_references')

Folder.Folder.get_references = get_references_nothing
Asset.Asset.get_references = get_references_nothing


def Ghost_get_references(self, version='public'):
    version = get_version_for_state(self, version)
    if version is None:
        return ()
    else:
        return [ ReferenceInfo(self, version._content_path,
                               ref_type=ReferenceInfo.REF_INTERN,
                               path_info=ReferenceInfo.PATH_ZOPE_ROOT,
                               restriction=ReferenceInfo.RESTRICT_CONTENT) ]


Ghost.Ghost.get_references = Ghost_get_references


def SQLDataSource_get_references(self, version=None):

    # check reference by ourselves, as the reference-checker
    # is not flexible enough :(
    status = ReferenceInfo.STATUS_OK
    connection_id = str(self.connection_id())
    try:
        o = getattr(self, connection_id)
        o = aq_base(o)
        # check if this is really a sql connection ... hmmm
        # cut & paste from the ZSQLMethods product :-/
        if hasattr(o,'_isAnSQLConnection') and o._isAnSQLConnection \
           and hasattr(o,'id'):
            # ok, if seems to be a connection
            pass
        else:
            status = ReferenceInfo.STATUS_WRONG_TYPE
    except (AttributeError, KeyError):
        status = ReferenceInfo.STATUS_MISSING

    return [ ReferenceInfo(self, connection_id,
                               ref_type=ReferenceInfo.REF_INTERN,
                               path_info=ReferenceInfo.PATH_OBJECT,
                               status = status) ]

try:
    from Products.SilvaExternalSources.SQLSource import SQLSource
    SQLSource.get_references = SQLDataSource_get_references
except ImportError:
    pass

try:
    from Products.Silva.SQLDataSource import SQLDataSource
    SQLDataSource.get_references = SQLDataSource_get_references
except ImportError:
    pass


def xml_reference_helper(ref_list, node, check_map):

    #print "xml_reference_helper: %s" % node.nodeName
    
    if node.nodeType != node.ELEMENT_NODE:
        return

    for child in node.childNodes:
        xml_reference_helper(ref_list, child, check_map)

    name = node.nodeName

    c = check_map.get(name, None)
    if c is not None:
        ref = c(node)
        if ref is not None:
            if type(ref) == TupleType or  type(ref) == ListType:
                ref_list.extend(ref)
            else:
                ref_list.append(ref)


def check_link(node):
    link = node.getAttribute('url')
    return ReferenceInfo(node.get_content(), link)

def check_image(node):
    path = node.getAttribute('path')
    link = node.getAttribute('link')

    # XXX should this be necesary ?
    if path and path[:1]=='/':
        path_info = ReferenceInfo.PATH_ZOPE_ROOT
    else:
        path_info = ReferenceInfo.PATH_CONTAINER
    path_ref = ReferenceInfo(node.get_content(), path,
                             ref_type = ReferenceInfo.REF_INTERN,
                             restriction=ReferenceInfo.RESTRICT_IMAGE,
                             path_info = path_info)
    refs = [ path_ref ]
    if link:
        link_ref = ReferenceInfo(node.get_content(), link)
        refs.append(link_ref)
    return refs


def check_file(node):
    # unused (except for the SilvaSampleDocument.SampleDocument ;-)
    path = node.getAttribute('path')

    return ReferenceInfo(node.get_content(), path,
                         ref_type = ReferenceInfo.REF_INTERN,
                         restriction = ReferenceInfo.RESTRICT_FILE,
                         path_info = ReferenceInfo.PATH_SILVA_ROOT)
    

def check_code(node):
    path = node.getAttribute('path')

    return ReferenceInfo(node.get_content(), path,
                         ref_type = ReferenceInfo.REF_INTERN,
                         restriction = ReferenceInfo.RESTRICT_CODE,
                         path_info = ReferenceInfo.PATH_ZOPE_ROOT)

        
def check_external_data(node):
    path = node.getAttribute('path')

    # hm: data source restriction ?
    return ReferenceInfo(node.get_content(), path,
                         ref_type = ReferenceInfo.REF_INTERN,
                         restriction = ReferenceInfo.RESTRICT_EXTSOURCE,
                         path_info = ReferenceInfo.PATH_ZOPE_ROOT)


default_check_map = {
    'link' : check_link,
    'file' : check_file,
    'image': check_image,
    'code' : check_code,
    'externaldata' : check_external_data,
    }
    

def Document_get_references(self, version='public'):
    version = get_version_for_state(self, version)
    if version is None:
        return ()

    if hasattr(aq_base(version),'content'):
        version = version.content
    
    refs = []
    xml_reference_helper(refs, version.documentElement, default_check_map)
    return refs


try:
    from Products.SilvaDocument.Document import Document
    Document.get_references = Document_get_references
except ImportError:
    try:
        from Products.Silva.Document import Document
        Document.get_references = Document_get_references
    except ImportError:
        pass

try:
    from Products.Silva.DemoObject import DemoObject
    DemoObject.get_references = Document_get_references
except ImportError:
    pass

try:
    from Products.SilvaDemoExtension.DemoExtensionObject import DemoExtensionObject
    DemoExtensionObject.get_references = Document_get_references
except ImportError:
    pass


zLOG.LOG('Silva Reference Checker', -100, \
         'monkey patched Silva core content types')
